use std::collections::VecDeque;
use std::io::ErrorKind;
use std::marker::PhantomData;
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::{Arc, Mutex};
use std::thread::JoinHandle;
use std::time::Duration;

use bevy::ecs::event::Events;
use bevy::prelude::{App, Commands, CoreStage, EventReader, Plugin, ResMut};
use thiserror::Error;
use websocket::{Message, OwnedMessage, WebSocketError};

use crate::events::{
	FromSocketMessage, RemoteEventResult, SocketControlEvent, SocketState, ToSocketMessage,
};

fn message_stack<T>() -> Arc<Mutex<VecDeque<T>>> {
	Arc::new(Mutex::new(VecDeque::<T>::new()))
}

pub struct RemoteEventDispatcher<SentMessageType, ReceivedMessageType>
where
	ReceivedMessageType: FromSocketMessage + Send + Sync + 'static,
	SentMessageType: ToSocketMessage + Send + Sync + 'static,
{
	pub status: SocketState,
	received_events: Arc<Mutex<VecDeque<ReceivedMessageType>>>,
	sent_events: Arc<Mutex<VecDeque<SentMessageType>>>,
	control_events: Arc<Mutex<VecDeque<SocketControlEvent>>>,
	receiver_handle: JoinHandle<()>,
	sender_handle: JoinHandle<()>,
	kill_sig: Arc<AtomicBool>,
}

impl<SentMessageType, ReceivedMessageType>
	RemoteEventDispatcher<SentMessageType, ReceivedMessageType>
where
	ReceivedMessageType: FromSocketMessage + Send + Sync + 'static,
	SentMessageType: ToSocketMessage + Send + Sync + 'static,
{
	pub fn shutdown(&mut self) {
		self.kill_sig.store(true, Ordering::Relaxed);
	}
}

#[derive(Error, Debug)]
pub enum RemoteEventError {
	#[error("{0}")]
	ConnectionStringError(#[from] websocket::client::ParseError),
	#[error("{0}")]
	WebsocketError(#[from] websocket::WebSocketError),
	#[error("{0}")]
	IOError(#[from] std::io::Error),
}

fn connect_to<SentMessageType, ReceivedMessageType, T: ToString>(
	remote_host: T,
) -> RemoteEventResult<RemoteEventDispatcher<SentMessageType, ReceivedMessageType>>
where
	ReceivedMessageType: FromSocketMessage + Send + Sync + 'static,
	SentMessageType: ToSocketMessage + Send + Sync + 'static,
{
	let client = websocket::ClientBuilder::new(&remote_host.to_string())?.connect_insecure()?;
	let sent_events = message_stack::<SentMessageType>();
	let received_events = message_stack();
	let control_events = message_stack();

	let mut kill_sig = Arc::new(AtomicBool::new(false));

	let (mut recv_channel, mut send_channel) = client.split()?;

	let mut recv = received_events.clone();
	let mut recv_control = control_events.clone();
	let receiver_handle = std::thread::spawn(move || {
		for event in recv_channel.incoming_messages() {
			match event {
				Ok(OwnedMessage::Text(string)) => recv
					.lock()
					.unwrap()
					.push_back(ReceivedMessageType::from_text(string)),
				Ok(OwnedMessage::Binary(bin)) => recv
					.lock()
					.unwrap()
					.push_back(ReceivedMessageType::from_binary(bin)),
				Err(e) => {
					match e {
						WebSocketError::NoDataAvailable => {}
						_ => log::error!("Socket Error: {}", e),
					}
					break;
				}
				_ => {}
			}
		}
		if let Err(ref err) = recv_channel.shutdown_all() {
			match &err.kind() {
				ErrorKind::NotConnected => {}
				_ => log::error!("{}", err),
			}
		}

		recv_control
			.lock()
			.unwrap()
			.push_front(SocketControlEvent::Disconnect);
	});

	let mut send = sent_events.clone();
	let mut send_control = control_events.clone();
	let send_kill_sig = kill_sig.clone();
	let sender_handle = std::thread::spawn(move || loop {
		{
			let mut lock = send.lock().unwrap();
			while let Some(to_send) = lock.pop_front() {
				if let Err(e) = send_channel.send_message(&Message::text(to_send.to_text())) {
					log::error!("{}", e);
				};
			}
		};

		if send_kill_sig.load(Ordering::Relaxed) {
			if let Err(ref err) = send_channel.shutdown_all() {
				match &err.kind() {
					ErrorKind::NotConnected => {}
					_ => log::error!("{}", err),
				}
			}

			send_control
				.lock()
				.unwrap()
				.push_front(SocketControlEvent::Disconnect);
		}
		std::thread::sleep(Duration::from_millis(75));
	});

	Ok(RemoteEventDispatcher {
		status: SocketState::Connected,
		sent_events,
		received_events,
		control_events,
		sender_handle,
		receiver_handle,
		kill_sig,
	})
}

pub fn socket_manager_system<SentMessageType, ReceivedMessageType>(
	mut commands: Commands,
	mut maybe_client: Option<ResMut<RemoteEventDispatcher<SentMessageType, ReceivedMessageType>>>,
	mut control_events: EventReader<SocketControlEvent>,
) where
	ReceivedMessageType: FromSocketMessage + Send + Sync + 'static,
	SentMessageType: ToSocketMessage + Send + Sync + 'static,
{
	if let Some(ref mut client) = maybe_client {
		let mut reboot_client = None;
		if let Ok(mut control_events) = client.control_events.lock() {
			while let Some(event) = control_events.pop_front() {
				match event {
					SocketControlEvent::Disconnect => {
						#[rustfmt::skip]
						// rustfmt breaks the ending "();" to put ");" on its own line...
						commands.remove_resource::<RemoteEventDispatcher<SentMessageType, ReceivedMessageType>>();
					}
					SocketControlEvent::Create(url) => {
						reboot_client = Some(url);
					}
				}
			}
		}

		if let Some(url) = reboot_client {
			client.shutdown();
			match connect_to::<SentMessageType, ReceivedMessageType, String>(url.to_string()) {
				Ok(client) => {
					commands.insert_resource(client);
				}
				Err(e) => {
					log::error!("Error creating socket connection: {}", e);
				}
			}
		}
	}

	for event in control_events.iter() {
		match event {
			SocketControlEvent::Create(url) => {
				if maybe_client.is_some() {
					log::warn!(
						"Trying to create a second socket connection while the first is alive"
					)
				} else {
					match connect_to::<SentMessageType, ReceivedMessageType, String>(
						url.to_string(),
					) {
						Ok(client) => {
							commands.insert_resource(client);
						}
						Err(e) => {
							log::error!("Error creating socket connection: {}", e);
						}
					}
				}
			}
			SocketControlEvent::Disconnect => match maybe_client {
				Some(ref mut client) => {
					client.shutdown();
					#[rustfmt::skip]
					// rustfmt breaks the ending "();" to put ");" on its own line...
					commands.remove_resource::<RemoteEventDispatcher<SentMessageType, ReceivedMessageType>>();
				}
				None => {
					log::warn!("There was no connection to disconnect from");
				}
			},
		}
	}
}

pub fn socket_event_wrangler<SentMessageType, ReceivedMessageType>(
	mut client: Option<ResMut<RemoteEventDispatcher<SentMessageType, ReceivedMessageType>>>,
	mut received_batch: ResMut<Events<ReceivedMessageType>>,
	mut sent_batch: ResMut<Events<SentMessageType>>,
) where
	ReceivedMessageType: FromSocketMessage + Send + Sync + 'static,
	SentMessageType: ToSocketMessage + Send + Sync + 'static,
{
	if let Some(mut client) = client {
		if let Ok(mut list) = client.sent_events.lock() {
			for message in sent_batch.drain() {
				list.push_back(message);
			}
		}

		let mut lock = client.received_events.lock().unwrap();
		while let Some(message) = lock.pop_front() {
			received_batch.send(message);
		}
	}
}

#[derive(Default)]
pub struct RemoteEventPlugin<SentMessageType, ReceivedMessageType>
where
	ReceivedMessageType: FromSocketMessage + Send + Sync + 'static,
	SentMessageType: ToSocketMessage + Send + Sync + 'static,
{
	_send: PhantomData<SentMessageType>,
	_recv: PhantomData<ReceivedMessageType>,
}

impl<SentMessageType, ReceivedMessageType> RemoteEventPlugin<SentMessageType, ReceivedMessageType>
where
	ReceivedMessageType: FromSocketMessage + Send + Sync + 'static,
	SentMessageType: ToSocketMessage + Send + Sync + 'static,
{
	pub fn new() -> Self {
		Self {
			_recv: Default::default(),
			_send: Default::default(),
		}
	}
}

impl<SentMessageType, ReceivedMessageType> Plugin
	for RemoteEventPlugin<SentMessageType, ReceivedMessageType>
where
	ReceivedMessageType: FromSocketMessage + Send + Sync + 'static,
	SentMessageType: ToSocketMessage + Send + Sync + 'static,
{
	fn build(&self, app: &mut App) {
		app.add_event::<SentMessageType>()
			.add_event::<ReceivedMessageType>()
			.add_event::<SocketControlEvent>()
			.add_system_to_stage(
				CoreStage::Last,
				socket_manager_system::<SentMessageType, ReceivedMessageType>,
			)
			.add_system_to_stage(
				CoreStage::PostUpdate,
				socket_event_wrangler::<SentMessageType, ReceivedMessageType>,
			);
	}
}
